/* NSC_LIC_BSD
 *
 * Revision 2.1.2
 *
 * National Semiconductor Corporation Open Source License for 
 *
 *    Geode Linux ALSA audio driver
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c) July, 2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 *   * Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer. 
 *
 *   * Redistributions in binary form must reproduce the above 
 *     copyright notice, this list of conditions and the following 
 *     disclaimer in the documentation and/or other materials provided 
 *     with the distribution. 
 *
 *   * Neither the name of the National Semiconductor Corporation nor 
 *     the names of its contributors may be used to endorse or promote 
 *     products derived from this software without specific prior 
 *     written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER 
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN 
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF 
 * YOUR JURISDICTION. It is licensee's responsibility to comply with 
 * any export regulations applicable in licensee's jurisdiction. Under 
 * CURRENT (July, 2001) U.S. export regulations this software 
 * is eligible for export from the U.S. and can be downloaded by or 
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed 
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran, 
 * Syria, Sudan, Afghanistan and any other country to which the U.S. 
 * has embargoed goods and services. 
 *
 * END_NSC_LIC_BSD */

 
#define SND_MAIN_OBJECT_FILE 
#include "../include/driver.h" 
#include "../include/initval.h" 
 
#include "../include/geode.h" 
 
MODULE_DESCRIPTION("Driver: NSC Geode\nCard: Generic NSC Geode Chipset\n"); 
 
char *snd_id[SND_CARDS] = SND_DEFAULT_STR;	/* ID for this card */ 
 
int snd_index[SND_CARDS]			= SND_DEFAULT_IDX;	/* Index 1-MAX */ 
int snd_dac_frame_size[SND_CARDS]	= {[0 ... (SND_CARDS - 1)] = 32 }; 
int snd_adc_frame_size[SND_CARDS]	= {[0 ... (SND_CARDS - 1)] = 32 }; 
int snd_gesbuf[SND_CARDS]			= {[0 ... (SND_CARDS - 1)] = 1024 }; 
int snd_midi_enable[SND_CARDS]		= {[0 ... (SND_CARDS - 1)] = 0 }; 
int snd_pcm_channels_p[SND_CARDS]	= {[0 ... (SND_CARDS - 1)] = 1 };   //   This is the number of Simultaneous streams 
int snd_pcm_channels_c[SND_CARDS]	= {[0 ... (SND_CARDS - 1)] = 1 };	//   This is the number of Simultaneous streams 
 
MODULE_PARM     (snd_index, "1-" __MODULE_STRING(SND_CARDS) "i"); 
MODULE_PARM_DESC(snd_index, "Index value for NSC Geode audio Controller."); 
 
MODULE_PARM		(snd_id, "1-" __MODULE_STRING(SND_CARDS) "s"); 
MODULE_PARM_DESC(snd_id, "ID string for NSC Geode audio Controller."); 
 
MODULE_PARM		(snd_gesbuf, "1-" __MODULE_STRING(SND_CARDS) "i"); 
MODULE_PARM_DESC(snd_gesbuf, "total Buffer Size in kB. [range=1-4096]"); 
 
MODULE_PARM		(snd_dac_frame_size, "1-" __MODULE_STRING(SND_CARDS) "i"); 
MODULE_PARM_DESC(snd_dac_frame_size,"DAC frame size in kB for NSC Geode audio Controller. [range=1-32]"); 
 
MODULE_PARM		(snd_adc_frame_size, "1-" __MODULE_STRING(SND_CARDS) "i"); 
MODULE_PARM_DESC(snd_adc_frame_size, "ADC frame size in kB for NSC Geode audio Controller. [range=1-32]"); 
 
MODULE_PARM		(snd_midi_enable, "1-" __MODULE_STRING(SND_CARDS) "i"); 
MODULE_PARM_DESC(snd_midi_enable, "Midi enabled for NSC Geode audio Controller. [BOOL]"); 
 
MODULE_PARM		(snd_pcm_channels_p, "1-" __MODULE_STRING(SND_CARDS) "i"); 
MODULE_PARM_DESC(snd_pcm_channels_p, "PCM Playback subchannels for NSC Geode audio Controller. [range=1-8]"); 
 
MODULE_PARM		(snd_pcm_channels_c, "1-" __MODULE_STRING(SND_CARDS) "i"); 
MODULE_PARM_DESC(snd_pcm_channels_c, "PCM Capture subchannels for NSC Geode audio Controller. [range=0-8]"); 
 
struct snd_geode { 
	struct pci_dev *pci; 
	snd_irq_t *irqptr; 
	snd_dma_t *dma1ptr; 
	snd_card_t *card; 
	geode_t *geode; 
	snd_pcm_t *pcm; 
	snd_kmixer_t *mixer; 
	snd_rawmidi_t *midi_uart; 
	int type; 
}; 
 
static struct snd_geode *snd_geode_cards[SND_CARDS] = SND_DEFAULT_PTR; 
 
static void snd_geode_use_inc(snd_card_t * card) 
{ 
	MOD_INC_USE_COUNT; 
} 
 
static void snd_geode_use_dec(snd_card_t * card) 
{ 
	MOD_DEC_USE_COUNT; 
} 
 
static int __init snd_geode_detect(snd_card_t * card,  
								   struct snd_geode *scard, 
								   unsigned short vendor,  
								   unsigned short device) 
{ 
	//snd_printk("Checking device %04X\n", device); 
	if ((scard->pci = pci_find_device(vendor,device,scard->pci)) == NULL) 
		    return -ENODEV; 
 
	return 0; 
} 
 
static void snd_geode_interrupt2(int irq, void *dev_id, struct pt_regs *regs) 
{ 
	struct snd_geode *scard = (struct snd_geode *) dev_id; 
 
	if (!scard || !scard->geode) 
		return; 
 
	snd_geode_interrupt(scard->geode); 
} 
 
static int __init snd_geode_resources(snd_card_t * card, struct snd_geode *scard, int dev) 
{ 
	int err; 
 
	// Registering DMA & IRQ  
	if ((err = snd_register_interrupt(	card,  
										"NSC Geode",  
										5,							// IRQ is fixed 
										SND_IRQ_TYPE_PCI,  
										snd_geode_interrupt2, 
										scard,  
										NULL,  
										&scard->irqptr 
										)) < 0) 
		return err; 
 
	if ((err = 
	     snd_register_dma_channel(card, "NSC Geode - DAC Frame", 0, 
				      SND_DMA_TYPE_PCI, 
				      snd_dac_frame_size[dev], NULL, 
				      &scard->dma1ptr)) < 0) 
		return err; 
 
	scard->dma1ptr->multi = 1;			// Enable multi-alloc  
    scard->dma1ptr->addressbits = 28;	// 256Mb  
 
	return 0; 
} 
 
static int __init snd_geode_probe(int dev, struct snd_geode *scard) 
{ 
	snd_card_t *card; 
	snd_kmixer_t *mixer = NULL; 
	snd_pcm_t *pcm = NULL; 
	snd_rawmidi_t *midi_uart = NULL; 
	int ret=0; 
 
	card = snd_card_new(snd_index[dev],  
						snd_id[dev],  
						snd_geode_use_inc,  
						snd_geode_use_dec 
						); 
 
	if (!card) 
		return -ENOMEM; 
 
	snd_printk("Looking for the Controller...\n"); 

	if (!snd_geode_detect(card, scard, PCI_VENDOR_ID_CYRIX, PCI_DEVICE_ID_UNICORN))  
	{ 
			snd_printk("Found: NSC UNICORN Audio Controller\n"); 
	} 
	else
	{
		if (!snd_geode_detect(card, scard, PCI_VENDOR_ID_CYRIX, PCI_DEVICE_ID_CENTAURUS))  
		{ 
				snd_printk("Found: NSC CENTAURUS Audio Controller\n"); 
		} 
		else
		{
			if (!snd_geode_detect(card, scard, PCI_VENDOR_ID_NSC, PCI_DEVICE_ID_CENTAURUS))  
			{ 
					snd_printk("Found: NSC CENTAURUS Audio Controller\n"); 
			} 
			else
			{
				if (!snd_geode_detect(card, scard, PCI_VENDOR_ID_NSC, PCI_DEVICE_ID_SCORPIUS))  
				{ 
						snd_printk("Found: NSC SCORPIUS Audio Controller\n"); 
				} 
				else
				{
					if (!snd_geode_detect(card, scard, PCI_VENDOR_ID_NSC, PCI_DEVICE_ID_SC1200))  
					{ 
							snd_printk("Found: NSC SC1200 Audio Controller\n"); 
					} 
					else
					{
						if (!snd_geode_detect(card, scard, PCI_VENDOR_ID_NSC, PCI_DEVICE_ID_CS5535))  
						{ 
								snd_printk("Found: 5535 - Capistrano Audio Controller\n"); 
						} 
					}
				}
			}
		}
	}
 
	if (scard->pci == NULL)  
	{ 
		snd_printk("Didn't find any NSC Controllers - ABORTING\n"); 
		snd_card_free(card); 
		return -ENODEV; 
	} 
 
	if (snd_geode_resources(card, scard, dev) < 0)  
	{ 
		snd_printk("No resources - ABORTING\n"); 
		snd_card_free(card); 
		return -ENODEV; 
	} 
 
	if (snd_geode_create(card, 
						 scard->pci,  
						 scard->dma1ptr,  
						 scard->irqptr, 
                         snd_midi_enable[dev],  
						 snd_gesbuf[dev], 
                         snd_pcm_channels_p[dev],  
						 snd_pcm_channels_c[dev],  
						 &scard->geode)   
						 < 0) 
		goto __nodev; 
 
	if (snd_geode_pcm(scard->geode, 0, &pcm) < 0) 
		goto __nodev; 
 
	if (snd_geode_mixer(scard->geode, 0, pcm, &mixer) < 0) 
	{ 
		snd_printk("*** snd_geode_mixer FAILED\n"); 
		goto __nodev; 
	} 
 
	strcpy(card->abbreviation, "GEODE"); 
	strcpy(card->shortname, "NSC Geode"); 
	sprintf(card->longname, "%s at 0x%lx, irq %li",	card->shortname, scard->geode->io_port, scard->irqptr->irq); 
 
	if (!snd_card_register(card)) 
	{ 
		scard->card = card; 
		scard->mixer = mixer; 
		scard->pcm = pcm; 
		scard->midi_uart = scard->geode->rmidi = midi_uart; 
		return 0; 
	} 
	else 
		snd_printk("*** snd_card_register FAILED - ret=%08X\n", ret); 
 
	goto __nodev; 
 
      __nodev: 
	snd_card_free(card); 
	return -ENXIO; 
} 
 
static int __exit snd_geode_free2(int dev) 
{ 
	struct snd_geode *scard; 
 
	scard = snd_geode_cards[dev]; 
	snd_geode_cards[dev] = NULL; 
	 
	if (scard)  
	{ 
		snd_card_unregister(scard->card); 
		snd_kfree(scard); 
	} 
 
	return 0; 
} 
 
static int __init alsa_card_geode_init(void) 
{ 
	int dev=0, cards=0; 
	struct snd_geode *scard; 
 
	scard = (struct snd_geode *) snd_kcalloc(sizeof(struct snd_geode), GFP_KERNEL); 

	if (snd_geode_probe(0, scard) < 0)  
	{ 
		snd_kfree(scard); 
	} 

	snd_geode_cards[0] = scard; 
 
	if (!scard)  
	{ 
		snd_printk("NSC Geode audio Controller #%i not found or device busy\n", dev + 1); 
		return -ENODEV; 
	} 

	return 0; 
} 
 
static void __exit alsa_card_geode_exit(void) 
{ 
	int dev; 
 
	for (dev = 0; dev < SND_CARDS; dev++) 
		snd_geode_free2(dev); 
} 
 
module_init(alsa_card_geode_init) 
module_exit(alsa_card_geode_exit) 
 
